#include "config.h"

#ifdef USE_GRAPHIC
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "const.h"
#include "nlllib.h"
#include "memory.h"
#include "image.h"

static image_t images[GRAPHIC_IMAGE_MAXNUM];
static int image_num = 0;

int image_init(void)
{
  image_num = 0;
  memset(images, 0, sizeof(images));
  return 0;
}

int image_done(void)
{
  int index;
  image_t image;

  for (index = 0; index < GRAPHIC_IMAGE_MAXNUM; index++) {
    image = images[index];
    if (image)
      image_destroy(image);
  }

  return 0;
}

int image_check(void)
{
  int index;

  if (image_num)
    return NLL_ERRCODE_MEMORY_NOT_EMPTY;

  for (index = 0; index < GRAPHIC_IMAGE_MAXNUM; index++) {
    if (images[index])
      return NLL_ERRCODE_MEMORY_NOT_EMPTY;
  }

  return 0;
}

static image_t _alloc(void)
{
  image_t image;
  if ((image = memory_alloc(sizeof(*image))) == NULL)
    return NULL;
  return image;
}

static int _free(image_t image)
{
  memory_free(image);
  return 0;
}

int image_is_draw_line(image_t image, image_draw_count_t count, int y)
{
  return (image->draw.line[y] >= count) ? 1 : 0;
}

int image_is_draw_column(image_t image, image_draw_count_t count, int x, int y)
{
  int i;
  i = image->draw.column_num * y + (x / DRAW_COLUMN_PER);
  return (image->draw.column[i] >= count) ? 1 : 0;
}

int image_is_draw_pixel(image_t image, image_draw_count_t count, int x, int y)
{
  return (image->draw.pixel[image->width * y + x] >= count) ? 1 : 0;
}

int image_set_draw_line(image_t image, image_draw_count_t count, int y)
{
  image->draw.line[y] = count;
  return 0;
}

int image_set_draw_column(image_t image, image_draw_count_t count, int x, int y)
{
  int i;
  i = image->draw.column_num * y + (x / DRAW_COLUMN_PER);
  image->draw.column[i] = count;
  return 0;
}

int image_set_draw_pixel(image_t image, image_draw_count_t count, int x, int y)
{
  image->draw.pixel[image->width * y + x] = count;
  return 0;
}

int image_get_color(image_t image, int x, int y)
{
  if ((x < 0) || (x > image->width  - 1) ||
      (y < 0) || (y > image->height - 1))
    return -1;
  return image->bitmap[image->width * y + x];
}

int image_set_color(image_t image, int x, int y, int color, unsigned int flags)
{
  if ((x < 0) || (x > image->width  - 1) ||
      (y < 0) || (y > image->height - 1))
    return -1;

  switch (flags & NLL_G_MODE_MASK) {
  case NLL_G_AND: color &= image_get_color(image, x, y); break;
  case NLL_G_OR:  color |= image_get_color(image, x, y); break;
  case NLL_G_XOR: color ^= image_get_color(image, x, y); break;
  case NLL_G_SET:
  default:
    break;
  }

  image->bitmap[image->width * y + x] = color;

  return 0;
}

int image_copy_save(image_t image, int x, int y, int width, int height)
{
  int i, n;

  if (width  < 0) { x += width;  width  = -width;  }
  if (height < 0) { y += height; height = -height; }

  if (x < 0) { width  += x; x = 0; }
  if (y < 0) { height += y; y = 0; }

  if (x + width  > image->width ) width  = image->width  - x;
  if (y + height > image->height) height = image->height - y;

  if ((width <= 0) || (height <= 0))
    return 0;

  for (i = 0; i < height; i++) {
    n = image->width * (y + i) + x;
    memcpy(&image->copy.bitmap[n], &image->bitmap[n],
	   image->width * sizeof(*(image->bitmap)));
  }

  return 0;
}

int image_copy_get_color(image_t image, int x, int y)
{
  if ((x < 0) || (x > image->width  - 1) ||
      (y < 0) || (y > image->height - 1))
    return -1;
  return image->copy.bitmap[image->width * y + x];
}

int image_copy_save_color(image_t image, int x, int y)
{
  int n;
  if ((x < 0) || (x > image->width  - 1) ||
      (y < 0) || (y > image->height - 1))
    return -1;
  n = image->width * y + x;
  image->copy.bitmap[n] = image->bitmap[n];
  return 0;
}

int image_get_index(image_t image)
{
  if (!image)
    return -1;
  return image->index;
}

image_t image_get_image(int index)
{
  image_t image;

  if (index >= GRAPHIC_IMAGE_MAXNUM)
    return NULL;

  if (index < 0) {
    for (index = 0; index < GRAPHIC_IMAGE_MAXNUM; index++) {
      image = images[index];
      if (image && image->status == IMAGE_STATUS_FREE) {
	image_destroy(image);
	image = images[index];
      }
      if (!image)
	break;
    }
    if (index == GRAPHIC_IMAGE_MAXNUM)
      return NULL;

    if ((image = _alloc()) == NULL)
      return NULL;
    image_num++;

    memset(image, 0, sizeof(*image));
    image->status = IMAGE_STATUS_NONE;
    image->index = index;

    images[index] = image;
  }

  return images[index];
}

int image_reset_draw(image_t image)
{
  memset(image->draw.line, 0, image->height * sizeof(image_draw_count_t));
  memset(image->draw.column, 0, image->draw.column_num * image->height * sizeof(image_draw_count_t));
  memset(image->draw.pixel, 0, image->width * image->height * sizeof(image_draw_count_t));
  return 0;
}

int image_destroy(image_t image)
{
  int r;

  image->status = IMAGE_STATUS_NONE;

  images[image->index] = NULL;

  if (image->copy.bitmap) memory_free(image->copy.bitmap);
  if (image->draw.pixel)  memory_free(image->draw.pixel);
  if (image->draw.column) memory_free(image->draw.column);
  if (image->draw.line)   memory_free(image->draw.line);
  if (image->bitmap)      memory_free(image->bitmap);
  memset(image, 0, sizeof(*image));

  image_num--;
  if ((r = _free(image)) < 0)
    return r;

  return 0;
}

int image_create(image_t image, int width, int height, int *bitmap)
{
  image->width  = width;
  image->height = height;

  image->size = image->width * image->height * sizeof(*(image->bitmap));
  if (bitmap) {
    image->bitmap = bitmap;
  } else {
    image->bitmap = memory_alloc(image->size);
    if (!image->bitmap)
      goto err;
    memset(image->bitmap, 0, image->size);
  }

  image->draw.column_num = (image->width + DRAW_COLUMN_PER - 1) / DRAW_COLUMN_PER;

  image->draw.line   = memory_alloc(image->height * sizeof(image_draw_count_t));
  image->draw.column = memory_alloc(image->draw.column_num * image->height * sizeof(image_draw_count_t));
  image->draw.pixel  = memory_alloc(image->width * image->height * sizeof(image_draw_count_t));
  if (!image->draw.line || !image->draw.column || !image->draw.pixel)
    goto err;

  image->copy.bitmap = memory_alloc(image->size);
  if (!image->copy.bitmap)
    goto err;

  image_reset_draw(image);

  image->draw.count = 0;

  image->status = IMAGE_STATUS_ACTIVE;

  return 0;

err:
  image_destroy(image);
  return -1;
}

image_t image_get(int width, int height, int *bitmap)
{
  image_t image;

  image = image_get_image(-1);
  if (!image)
    return NULL;

  if (image_create(image, width, height, bitmap) < 0)
    return NULL;

  return image;
}
#endif
